﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Runtime.InteropServices;
using System.Runtime.InteropServices.ComTypes;

namespace clriecache
{
    class ClearIECache
    {
        [StructLayout(LayoutKind.Sequential)]
        internal struct INTERNET_CACHE_ENTRY_INFO
        {
            public UInt32 dwStructSize;
            public IntPtr lpszSourceUrlName;
            public IntPtr lpszLocalFileName;
            public UInt32 CacheEntryType;
            public UInt32 dwUseCount;
            public UInt32 dwHitRate;
            public UInt32 dwSizeLow;
            public UInt32 dwSizeHigh;
            public System.Runtime.InteropServices.ComTypes.FILETIME LastModifiedTime;
            public System.Runtime.InteropServices.ComTypes.FILETIME ExpireTime;
            public System.Runtime.InteropServices.ComTypes.FILETIME LastAccessTime;
            public System.Runtime.InteropServices.ComTypes.FILETIME LastSyncTime;
            public IntPtr lpHeaderInfo;
            public UInt32 dwHeaderInfoSize;
            public IntPtr lpszFileExtension;
            public UInt32 dwExemptDelta;
        }

        // For PInvoke: Initiates the enumeration of the cache groups in the Internet cache
        [DllImport("wininet.dll",
            SetLastError = true,
            CharSet = CharSet.Auto,
            CallingConvention = CallingConvention.StdCall)]
        internal static extern IntPtr FindFirstUrlCacheGroup(
            int dwFlags,
            int dwFilter,
            IntPtr lpSearchCondition,
            int dwSearchCondition,
            ref long lpGroupId,
            IntPtr lpReserved);

        // For PInvoke: Retrieves the next cache group in a cache group enumeration
        [DllImport("wininet.dll",
            SetLastError = true,
            CharSet = CharSet.Auto,
            CallingConvention = CallingConvention.StdCall)]
        [return: MarshalAs(UnmanagedType.Bool)]
        internal static extern bool FindNextUrlCacheGroup(
            IntPtr hFind,
            ref long lpGroupId,
            IntPtr lpReserved);

        // For PInvoke: Releases the specified GROUPID and any associated state in the cache index file
        [DllImport("wininet.dll",
            SetLastError = true,
            CharSet = CharSet.Auto,
            CallingConvention = CallingConvention.StdCall)]
        [return: MarshalAs(UnmanagedType.Bool)]
        internal static extern bool DeleteUrlCacheGroup(
            long GroupId,
            int dwFlags,
            IntPtr lpReserved);

        [DllImport("wininet.dll",
            SetLastError = true,
            CallingConvention = CallingConvention.StdCall)]
        [return: MarshalAs(UnmanagedType.Bool)]
        internal static extern bool FindCloseUrlCache(IntPtr ptrCacheEntryInfo);

        // For PInvoke: Begins the enumeration of the Internet cache
        [DllImport("wininet.dll",
            SetLastError = true,
            CharSet = CharSet.Auto,
            CallingConvention = CallingConvention.StdCall)]
        internal static extern IntPtr FindFirstUrlCacheEntry(
            [MarshalAs(UnmanagedType.LPTStr)] string lpszUrlSearchPattern,
            IntPtr lpFirstCacheEntryInfo,
            ref int lpdwFirstCacheEntryInfoBufferSize);

        // For PInvoke: Retrieves the next entry in the Internet cache
        [DllImport(@"wininet",
            SetLastError = true,
            CharSet = CharSet.Auto,
            CallingConvention = CallingConvention.StdCall)]
        [return: MarshalAs(UnmanagedType.Bool)]
        internal static extern bool FindNextUrlCacheEntry(
            IntPtr hFind,
            IntPtr lpNextCacheEntryInfo,
            ref int lpdwNextCacheEntryInfoBufferSize);

        // For PInvoke: Removes the file that is associated with the source name from the cache, if the file exists
        [DllImport(@"wininet",
            SetLastError = true,
            CharSet = CharSet.Auto,
            CallingConvention = CallingConvention.StdCall)]
        [return: MarshalAs(UnmanagedType.Bool)]
        internal static extern bool DeleteUrlCacheEntry(
            IntPtr lpszUrlName);

        // Indicates that all of the cache groups in the user's system should be enumerated
        const int CACHEGROUP_SEARCH_ALL = 0x0;

        // Indicates that all the cache entries that are associated with the cache group
        // should be deleted, unless the entry belongs to another cache group.
        const int CACHEGROUP_FLAG_FLUSHURL_ONDELETE = 0x2;

        // File not found.
        const int ERROR_FILE_NOT_FOUND = 0x2;

        // No more items have been found.
        const int ERROR_NO_MORE_ITEMS = 259;

        // Cookie cache entry.
        const int COOKIE_CACHE_ENTRY = 0x00100000;

        HashSet<string> cacheFolders = null;

        private void ClearCacheGroups(bool verbose, bool readOnly)
        {
            IntPtr enumHandle = IntPtr.Zero;

            // Pointer to a GROUPID variable
            long groupId = 0;

            bool returnValue = true;

            try
            {
                // Delete the groups first.
                // Groups may not always exist on the system.
                // For more information, visit the following Microsoft Web site:
                // http://msdn.microsoft.com/library/?url=/workshop/networking/wininet/overview/cache.asp			
                // By default, a URL does not belong to any group. Therefore, that cache may become
                // empty even when the CacheGroup APIs are not used because the existing URL does not belong to any group.			
                enumHandle = FindFirstUrlCacheGroup(0, CACHEGROUP_SEARCH_ALL, IntPtr.Zero, 0, ref groupId, IntPtr.Zero);

                // If there are no items in the Cache, you are finished.
                if (enumHandle != IntPtr.Zero && ERROR_NO_MORE_ITEMS == Marshal.GetLastWin32Error())
                    return;

                // Loop through Cache Group, and then delete entries.
                while (returnValue)
                {
                    if (ERROR_NO_MORE_ITEMS == Marshal.GetLastWin32Error() || ERROR_FILE_NOT_FOUND == Marshal.GetLastWin32Error())
                        break;

                    if (verbose)
                    {
                        Console.WriteLine("enumHandle=0x" + enumHandle.ToString("X8"));
                        Console.WriteLine("groupId=0x" + groupId.ToString("X8"));
                    }

                    // Delete a particular Cache Group.
                    if (!readOnly)
                        returnValue = DeleteUrlCacheGroup(groupId, CACHEGROUP_FLAG_FLUSHURL_ONDELETE, IntPtr.Zero);

                    if (returnValue || (!returnValue && ERROR_FILE_NOT_FOUND == Marshal.GetLastWin32Error()))
                    {
                        returnValue = FindNextUrlCacheGroup(enumHandle, ref groupId, IntPtr.Zero);
                    }
                }
            }
            finally
            {
                if (enumHandle != IntPtr.Zero)
                    FindCloseUrlCache(enumHandle);
            }
        }

        private void ClearCacheEntries(bool deleteCookies, bool verbose, bool readOnly)
        {
            IntPtr enumHandle = IntPtr.Zero;

            int cacheEntryInfoBufferSizeInitial = 0;
            int cacheEntryInfoBufferSize = 0;
            IntPtr cacheEntryInfoBuffer = IntPtr.Zero;
            INTERNET_CACHE_ENTRY_INFO internetCacheEntry;
            bool returnValue = true;

            try
            {
                // Start to delete URLs that do not belong to any group.
                // Detect cacheEntryInfoBufferSizeInitial size.
                enumHandle = FindFirstUrlCacheEntry(null, IntPtr.Zero, ref cacheEntryInfoBufferSizeInitial);

                if (enumHandle == IntPtr.Zero && ERROR_NO_MORE_ITEMS == Marshal.GetLastWin32Error())
                    return;

                cacheEntryInfoBufferSize = cacheEntryInfoBufferSizeInitial;
                cacheEntryInfoBuffer = Marshal.AllocHGlobal(cacheEntryInfoBufferSize);
                FindCloseUrlCache(enumHandle);
                enumHandle = FindFirstUrlCacheEntry(null, cacheEntryInfoBuffer, ref cacheEntryInfoBufferSizeInitial);

                while (returnValue)
                {
                    internetCacheEntry = (INTERNET_CACHE_ENTRY_INFO)Marshal.PtrToStructure(cacheEntryInfoBuffer, typeof(INTERNET_CACHE_ENTRY_INFO));

                    bool isCookie = (internetCacheEntry.CacheEntryType & COOKIE_CACHE_ENTRY) == COOKIE_CACHE_ENTRY;
                    string fileName = Marshal.PtrToStringAuto(internetCacheEntry.lpszLocalFileName);

                    if (verbose)
                    {
                        Console.WriteLine("-----------------------------");
                        Console.WriteLine("entry type=0x" + internetCacheEntry.CacheEntryType.ToString("X8"));
                        Console.WriteLine(Marshal.PtrToStringAuto(internetCacheEntry.lpszSourceUrlName));
                        Console.WriteLine(fileName);
                    }

                    //Console.WriteLine(fileName);
                    //Console.WriteLine("isCookie=" + isCookie);

                    if (!readOnly && (!isCookie || deleteCookies))
                    {
                        if (!String.IsNullOrEmpty(fileName))
                        {
                            string folder = Path.GetDirectoryName(fileName);

                            if (!String.IsNullOrEmpty(folder))
                                cacheFolders.Add(folder);
                        }

                        returnValue = DeleteUrlCacheEntry(internetCacheEntry.lpszSourceUrlName);
                    }
                    //Console.WriteLine("returnValue1=" + returnValue.ToString());

                    if (returnValue || ERROR_FILE_NOT_FOUND == Marshal.GetLastWin32Error())
                    {
                        cacheEntryInfoBufferSizeInitial = cacheEntryInfoBufferSize;
                        returnValue = FindNextUrlCacheEntry(enumHandle, cacheEntryInfoBuffer, ref cacheEntryInfoBufferSizeInitial);
                        //Console.WriteLine("returnValue2=" + returnValue.ToString());

                        if (!returnValue && cacheEntryInfoBufferSizeInitial > cacheEntryInfoBufferSize)
                        {
                            cacheEntryInfoBufferSize = cacheEntryInfoBufferSizeInitial;
                            cacheEntryInfoBuffer = Marshal.ReAllocHGlobal(cacheEntryInfoBuffer, (IntPtr)cacheEntryInfoBufferSize);
                            returnValue = FindNextUrlCacheEntry(enumHandle, cacheEntryInfoBuffer, ref cacheEntryInfoBufferSizeInitial);
                            //Console.WriteLine("returnValue3=" + returnValue.ToString());
                        }
                    }
                }
            }
            finally
            {
                if (enumHandle != IntPtr.Zero)
                    FindCloseUrlCache(enumHandle);

                if (cacheEntryInfoBuffer != IntPtr.Zero)
                    Marshal.FreeHGlobal(cacheEntryInfoBuffer);
            }
        }

        internal void ClearCache(bool deleteCookies, bool verbose, bool readOnly)
        {
            cacheFolders = new HashSet<string>();

            ClearCacheGroups(verbose, readOnly);
            ClearCacheEntries(deleteCookies, verbose, readOnly);

            HashSet<string> keepFiles = new HashSet<string>(StringComparer.CurrentCultureIgnoreCase);
            keepFiles.Add("desktop.ini");
            keepFiles.Add("index.dat");

            foreach (var folder in cacheFolders)
            {
                //Console.WriteLine("folder=" + folder);

                if (Directory.Exists(folder))
                {
                    foreach (var file in Directory.GetFiles(folder, "*", SearchOption.TopDirectoryOnly))
                    {
                        string fileName = Path.GetFileName(file);

                        if (!String.IsNullOrEmpty(fileName) && !keepFiles.Contains(fileName))
                        {
                            try
                            {
                                File.Delete(file);
                                //Console.WriteLine("Deleted file: " + file);
                            }
                            catch { }
                        }
                    }
                }
            }
        }
    }
}
